//********************************************************************
//*** POSIX/time.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: time.cc,v 1.13 2009-03-02 14:16:43 mws Exp $
//********************************************************************

#ifndef POSIX_TIME_HH
#include "time.hh"
#endif

#include <cmath>
#include <iomanip>
#include <iostream>

#ifndef INCLUDED_THREAD_LOCK_HH
#include "Thread/Lock.hh"
#endif

#ifndef INCLUDED_THREAD_SYNCHRONIZE_HH
#include "Thread/Synchronize.hh"
#endif

namespace fatalmind {
namespace POSIX {

template<class TM>
timespec<TM>::timespec()
{
	tv_sec = 0;
	tv_nsec = 0;
}

template<class TM>
timespec<TM>::timespec(float t)
{
	tv_sec = static_cast<time_t>(floor(t));
	tv_nsec = static_cast<long>(floor((t-tv_sec)*1000000000L));
}

template<class TM>
timespec<TM>::timespec(time_t sec, long nsec)
{
	tv_sec = sec;
	tv_nsec = nsec;
}

template<class TM>
timespec<TM>::timespec(::timespec t)
: ::timespec(t)
{
}

template<class TM>
timespec<TM>::timespec(const timespec<TM>& t)
: ::timespec(t)
{
}

template<class TM>
timespec<TM>::~timespec()
{
	
}

template<class TM>
timespec<TM>& timespec<TM>::operator+=(const timespec<TM>& rho)
{
	Synchronize2<lock_t, lock_t> sync(_lk, rho._lk);
	
	tv_nsec += rho.tv_nsec;
	tv_sec  += rho.tv_sec;
	if (tv_nsec >= 1000000000L) {
		tv_sec += tv_nsec / 1000000000L;
		tv_nsec %= 1000000000L;
	}
	return *this;
}

template<class TM>
timespec<TM>& timespec<TM>::operator-=(const timespec<TM>& rho)
{
	Synchronize2<lock_t, lock_t> sync(_lk, rho._lk);
	tv_nsec -= rho.tv_nsec;	
	tv_sec  -= rho.tv_sec;
	if (tv_nsec < 0) {
		tv_sec--;
		tv_nsec += 1000000000L;
	}
	return *this;
}

template<class TM>
timespec<TM>::operator float() const
{
	Synchronize<lock_t> sync(_lk);
	float rv(tv_nsec);
	rv /= 1000000000L;
	rv += tv_sec;
	return rv;
}

template<class TM>
std::ostream& operator<<(std::ostream& os, const timespec<TM>& ts)
{
	Synchronize<typename timespec<TM>::lock_t> sync(ts._lk);
	char fill = os.fill();
	return os << ts.tv_sec 
			<< '.' << std::setw(9) << std::setfill('0') << ts.tv_nsec
			<< std::setfill(fill);
}

// TODO:: critical for threading...
template<class TM>
bool operator<(const timespec<TM>& lho, const int rho) {
	Synchronize<typename timespec<TM>::lock_t> sync(lho._lk);
	return lho.tv_sec < rho;
}

} // namespace POSIX
} // namespace fatalmind

