//********************************************************************
//*** ResourcePool/SQL/SQLStatement.t.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLStatement.t.cc,v 1.10 2009-05-04 13:20:43 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */
#include <string>
#include "SQLStatement.hh"
#include "SQLException.hh"
#include "t/main.cc"


using fatalmind::SQL::SQLStatement;
using fatalmind::SQL::SQLException;

void INIT_SQLStatement()
{
}

void TEST_SQLStatment_Parser_simple() {
	std::string sql1("Select * from x where a = ?");
	SQLStatement p(sql1);

	assertEqual(p.getParamCount(), 1);
	assertEqual(p.getNameFor(0), "");
	assertEqual(p.getTypeFor(0), SQLStatement::Parameter::UNKNOWN);
	assertEqual(p.getSQLFragment(0), "Select * from x where a = ");
	assertEqual(p.getSQLFragment(1), "");
}

void TEST_SQLStatment_Parser_more() {
	std::string sql1("Select * from x where a = ? and b = ? and c=? and d = \"?\" and e = '?'");
	SQLStatement p(sql1);

	assertEqual(p.getParamCount(), 3);

	assertEqual(p.getNameFor(0), "");
	assertEqual(p.getTypeFor(0), SQLStatement::Parameter::UNKNOWN);
	assertEqual(p.getNameFor(1), "");
	assertEqual(p.getTypeFor(1), SQLStatement::Parameter::UNKNOWN);
	assertEqual(p.getNameFor(2), "");
	assertEqual(p.getTypeFor(2), SQLStatement::Parameter::UNKNOWN);

	assertEqual(p.getSQLFragment(0), "Select * from x where a = ");
	assertEqual(p.getSQLFragment(1), " and b = ");
	assertEqual(p.getSQLFragment(2), " and c=");
	assertEqual(p.getSQLFragment(3), " and d = \"?\" and e = '?'");
}

void TEST_SQLStatment_Parser_named() {
	std::string sql1("Select * from x where a = ?{a} and b = ?{b} and c=?{c} and d = \"?\" and e = '?'");
	SQLStatement p(sql1);

	assertEqual(p.getParamCount(), 3);

	assertEqual(p.getNameFor(0), "a");
	assertEqual(p.getTypeFor(0), SQLStatement::Parameter::UNKNOWN);
	assertEqual(p.getNameFor(1), "b");
	assertEqual(p.getTypeFor(1), SQLStatement::Parameter::UNKNOWN);
	assertEqual(p.getNameFor(2), "c");
	assertEqual(p.getTypeFor(2), SQLStatement::Parameter::UNKNOWN);

	assertEqual(p.getSQLFragment(0), "Select * from x where a = ");
	assertEqual(p.getSQLFragment(1), " and b = ");
	assertEqual(p.getSQLFragment(2), " and c=");
	assertEqual(p.getSQLFragment(3), " and d = \"?\" and e = '?'");
}

void TEST_SQLStatment_Parser_typed() {
	std::string sql1("Select * from x where a = ?{:INT} and b = ?{:CHAR} and c=?{:FLOAT} and d = \"?\" and e = '?'");
	SQLStatement p(sql1);

	assertEqual(p.getParamCount(), 3);

	assertEqual(p.getNameFor(0), "");
	assertEqual(p.getTypeFor(0), SQLStatement::Parameter::INT);
	assertEqual(p.getNameFor(1), "");
	assertEqual(p.getTypeFor(1), SQLStatement::Parameter::CHAR);
	assertEqual(p.getNameFor(2), "");
	assertEqual(p.getTypeFor(2), SQLStatement::Parameter::FLOAT);

	assertEqual(p.getSQLFragment(0), "Select * from x where a = ");
	assertEqual(p.getSQLFragment(1), " and b = ");
	assertEqual(p.getSQLFragment(2), " and c=");
	assertEqual(p.getSQLFragment(3), " and d = \"?\" and e = '?'");
}

void TEST_SQLStatment_Parser_namedandtyped() {
	std::string sql1("Select * from x where a = ?{parama:INT} and b = ?{paramb:CHAR} and c=?{paramc:FLOAT} and d = \"?\" and e = '?'");
	SQLStatement p(sql1);

	assertEqual(p.getParamCount(), 3);

	assertEqual(p.getNameFor(0), "parama");
	assertEqual(p.getTypeFor(0), SQLStatement::Parameter::INT);
	assertEqual(p.getNameFor(1), "paramb");
	assertEqual(p.getTypeFor(1), SQLStatement::Parameter::CHAR);
	assertEqual(p.getNameFor(2), "paramc");
	assertEqual(p.getTypeFor(2), SQLStatement::Parameter::FLOAT);

	assertEqual(p.getSQLFragment(0), "Select * from x where a = ");
	assertEqual(p.getSQLFragment(1), " and b = ");
	assertEqual(p.getSQLFragment(2), " and c=");
	assertEqual(p.getSQLFragment(3), " and d = \"?\" and e = '?'");
}

void TEST_SQLStatment_Parser_doublename() {
	// names can be used twice!
	std::string sql1("Select * from x where a = ?{parama:INT} and b = ?{paramb:CHAR} and c=?{parama:INT} and d = \"?\" and e = '?'");
	SQLStatement p(sql1);

	assertEqual(p.getParamCount(), 3);

	assertEqual(p.getNameFor(0), "parama");
	assertEqual(p.getTypeFor(0), SQLStatement::Parameter::INT);
	assertEqual(p.getNameFor(1), "paramb");
	assertEqual(p.getTypeFor(1), SQLStatement::Parameter::CHAR);
	assertEqual(p.getNameFor(2), "parama");
	assertEqual(p.getTypeFor(2), SQLStatement::Parameter::INT);

	assertEqual(p.getSQLFragment(0), "Select * from x where a = ");
	assertEqual(p.getSQLFragment(1), " and b = ");
	assertEqual(p.getSQLFragment(2), " and c=");
	assertEqual(p.getSQLFragment(3), " and d = \"?\" and e = '?'");
}
void TEST_SQLStatment_Parser_docexamples() {
	assertEqual(SQLStatement("select * from dual").getParamCount(), 0);
	assertEqual(SQLStatement("select ? from dual").getParamCount(), 1);
	assertEqual(SQLStatement("select ? from dual x where a = ?{a}").getParamCount(), 2);

	assertEqual(SQLStatement("select ? from dual").getNameFor(0), "");
	assertEqual(SQLStatement("select ? from x where a = ?{a}").getNameFor(1), "a");
}

void TEST_SQLStatment_Parser_invalidtype() {
	try {	
		std::string sql1("Select * from x where a = ?{parama:INT} and b = ?{paramb:CHAR} and c=?{parama:blah} and d = \"?\" and e = '?'");
		SQLStatement p(sql1);
		assertNotReach();
	} catch (SQLException&) {
	}
}

void TEST_SQLStatment_Parser_invalidmethods() {
	SQLStatement stm("Select * from x where a = ?{parama:INT} and b = ?{paramb:CHAR} and c=?{parama} and d = \"?\" and e = '?'");
	try {
		stm.getNameFor(10);
		assertNotReach();
	} catch (const std::out_of_range& e) {
	}
	try {
		stm.getTypeFor(10);
		assertNotReach();
	} catch (const std::out_of_range& e) {
	}
}

void TEST_PERF_SQLStatement_mixed() {
	for (int i = 0; i < 10000; ++i) {
		SQLStatement("Select * from x where a = ? and b = ?{paramb:CHAR} and c=?{parama} and d = \"?\" and e = '?'");
	}
}

void TEST_PERF_SQLStatement_concatFragments() {
	SQLStatement sql ("Select * from x where a = ? and b = ?{paramb:CHAR} and c=?{parama} and d = \"?\" and e = '?'");
	std::string empty;
	for (int t = 0; t < 100000; ++t) {
		empty.clear();
		int p =sql.getParamCount();
		for (int i = 0; i < p; ++i) {
			empty.append(sql.getSQLFragment(i));
			empty.append("?");
		}	
		empty.append(sql.getSQLFragment(p));
	}
}

void TEST_PERF_SQLStatement_copy() {
	SQLStatement sql ("Select * from x where a = ? and b = ?{paramb:CHAR} and c=?{parama} and d = \"?\" and e = '?'");
	for (int t = 0; t < 100000; ++t) {
		SQLStatement sql2(sql);
	}
}

void TEST_SQLStatement_reassign() {
	SQLStatement sql; // default ctor
	SQLStatement blah("select * from dual where x = ?");

	assertEqual(sql.getSQL(), "");
	assertEqual(sql.getParamCount(), 0);

	assertEqual(blah.getSQL(), "select * from dual where x = ?");
	assertEqual(blah.getParamCount(), 1);

	sql = blah;

	assertEqual(sql.getSQL(), "select * from dual where x = ?");
	assertEqual(sql.getParamCount(), 1);
}

void CLEANUP_SQLStatement() {
}
