//********************************************************************
//*** ResourcePool/SQL/Types.t.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: Types.t.cc,v 1.29 2009-05-02 22:17:15 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */


#include "t/main.hh"
#include "ResourcePool/IgnoreExceptionCommand.hh"

using namespace std;
using namespace fatalmind;

#include <memory>
#include <iostream>
#include <limits>


void test_Types_drop_all_tables(RP& p, RP::Commands::SQLBatch& bc) {
	RP::Commands::SQLExecute tdrop_str(p, "drop table fatalmind_str");
	RP::Commands::IgnoreException drop_str(tdrop_str);

	RP::Commands::SQLExecute tdrop_num(p, "drop table fatalmind_numeric");
	RP::Commands::IgnoreException drop_num(tdrop_num);

	RP::Commands::SQLExecute tdrop_num30(p, "drop table fatalmind_numeric30");
	RP::Commands::IgnoreException drop_num30(tdrop_num30);

	RP::Commands::SQLExecute tdrop_double(p, "drop table fatalmind_double");
	RP::Commands::IgnoreException drop_double(tdrop_double);

	RP::Commands::SQLExecute tdrop_lob(p, "drop table fatalmind_lob");
	RP::Commands::IgnoreException drop_lob(tdrop_lob);


	bc.add(drop_str);
	bc.add(drop_lob);
	bc.add(drop_num);
	bc.add(drop_num30);
	bc.add(drop_double);

//	p.execute(batch);
}

void test_Types_create_all_tables(RP& p, RP::Commands::SQLBatch& bc) {
	RP::Commands::SQLExecute create_num(p, "create table fatalmind_numeric (id numeric not null primary key, val numeric)");
	RP::Commands::SQLExecute create_double(p, "create table fatalmind_double (id numeric not null primary key, val  DOUBLE PRECISION)");
	RP::Commands::SQLExecute create_num30(p, "create table fatalmind_numeric30 (id numeric not null primary key, val numeric(30))");
	RP::Commands::SQLExecute create_str(p, getStringTable(p));
	RP::Commands::SQLExecute create_lob(p, getLobTable(p));

	bc.add(create_num);
	bc.add(create_double);
	bc.add(create_num30);
	bc.add(create_str);
	bc.add(create_lob);
}

void test_Types_delete_all_tables(RP& p, RP::Commands::SQLBatch& bc) {
	RP::Commands::SQLExecute delete_num(p, "delete from fatalmind_numeric");
	RP::Commands::SQLExecute delete_double(p, "delete from fatalmind_double");
	RP::Commands::SQLExecute delete_num30(p, "delete from fatalmind_numeric30");
	RP::Commands::SQLExecute delete_str(p, "delete from fatalmind_str");
	RP::Commands::SQLExecute delete_lob(p, "delete from fatalmind_lob");

	bc.add(delete_num);
	bc.add(delete_double);
	bc.add(delete_num30);
	bc.add(delete_str);
	bc.add(delete_lob);
}

void test_Types_delete_all_tables(RP& p) {
	RP::Commands::SQLBatch bc(p);
	test_Types_delete_all_tables(p, bc);
	p.execute(bc);
}

void test_Types_init(RP& p)
{
	RP::Commands::SQLBatch batch(p, true);
	
	test_Types_drop_all_tables(p, batch);
	test_Types_create_all_tables(p, batch);
	test_Types_delete_all_tables(p, batch);
	while (batch.getExecPosition() != batch.end()) {
		try {
			p.execute(batch);
		} catch (fatalmind::Exception e) {
			// std::cout << "got exception " << e.message() << std::endl;
		}
	}
}

template<typename T>
void test_int(RP& p, int& id)
{
	int mxid = id++;
	int miid = id++;

	RP::Commands::SQLExecute insert(p, "insert into fatalmind_numeric (id, val) values (?, ?)");
	T max_val = std::numeric_limits<T>::max();
	insert.bind(0, mxid);
	insert.bind(1, max_val);
	p.execute(insert);
	
	T min_val = std::numeric_limits<T>::min();
	insert.bind(0, miid);
	insert.bind(1, min_val);
	p.execute(insert);


	RP::Commands::SQLSelectRow select(p, "select val,val from fatalmind_numeric where id = ?");
	T sel_val;
	double sel_valdouble;
	select.bindout(0, sel_val);
	select.bindout(1, sel_valdouble);

	select.bind(0, mxid);
	p.execute(select);
	assertEqualT(T, sel_val, std::numeric_limits<T>::max());
	assertEqualT(double, sel_valdouble,  std::numeric_limits<T>::max());

	select.bind(0, miid);
	p.execute(select);
	assertEqualT(T, sel_val, std::numeric_limits<T>::min());
	assertEqualT(double, sel_valdouble,  std::numeric_limits<T>::min());

	RP::Commands::SQLSelectRow select2(p, "select val - 1 from fatalmind_numeric where id = ?");
	select2.bind(0, mxid);
	select2.bindout(0, sel_val);
	p.execute(select2);
	assertEqualT(T, sel_val, (std::numeric_limits<T>::max()-1));

	try {
		select2.bind(0, miid);
		p.execute(select2);
		std::cerr << "FAIL: signed: " << std::numeric_limits<T>::is_signed << " size: " << sizeof(T) 
			<< " id max " << mxid << "  id min " << miid << std::endl;
		assertNotReach();
	} catch (SQL::SQLException e) {
	}
}

void test_Types_int(RP& p)
{
	int id = 0;
	test_int<short int>(p, id);
	test_int<int>(p, id);
	test_int<long int>(p, id);

	test_int<unsigned short int>(p, id);
	test_int<unsigned int>(p, id);
	test_int<unsigned long int>(p, id);
}

template<typename T>
void test_float(RP& p, int& id) 
{
	int nulid = id++;
	int tetid = id++;
	int temtid = id++;
	int mtemtid = id++;
	T val;

	RP::Commands::SQLExecute insert(p, "insert into fatalmind_double (id, val) values (?, ?)");
	val = 0.0;
	insert.bind(0, nulid);
	insert.bind(1, val);
	p.execute(insert);
	
	insert.bind(0, tetid);
	val = 10E10;
	insert.bind(1, val);
	p.execute(insert);

	insert.bind(0, temtid);
	val = 10E-10;
	insert.bind(1, val);
	p.execute(insert);

	insert.bind(0, mtemtid);
	val = -10E-10;
	insert.bind(1, val);
	p.execute(insert);


	RP::Commands::SQLSelectRow select(p, "select val from fatalmind_double where id = ?");
	T sel_val;
	select.bindout(0, sel_val);
	select.bind(0, nulid);
	p.execute(select);
	assertEqualT(T, sel_val, 0.0);

	select.bind(0, tetid);
	p.execute(select);
	assertEqualT(T, sel_val, 100000000000.0);

	select.bind(0, temtid);
	p.execute(select);
	assertEqualT(T, sel_val, 0.000000001);

	select.bind(0, mtemtid);
	p.execute(select);
	assertEqualT(T, sel_val, -0.000000001);

	RP::Commands::SQLSelectRow select2(p, "select val+? from fatalmind_double where id = ?");
	select2.bind(0, std::numeric_limits<T>::epsilon());	
	select2.bind(1, nulid);
	select2.bindout(0, sel_val);
	p.execute(select2);
	assertEqualT(T, sel_val, std::numeric_limits<T>::epsilon());

	select2.bind(1, tetid);
	p.execute(select2);
	assertEqualT(T, sel_val, 100000000000.0 + std::numeric_limits<T>::epsilon());

	select2.bind(1, temtid);
	p.execute(select2);
	assertEqualT(T, sel_val, 0.000000001 + std::numeric_limits<T>::epsilon());

	select2.bind(1, mtemtid);
	p.execute(select2);
	assertEqualT(T, sel_val, -0.000000001 + std::numeric_limits<T>::epsilon());

}

void test_Types_float(RP& p)
{
	int id = 12;
	{
		test_Types_delete_all_tables(p);	
		test_float<float>(p, id);
	}
}
void test_Types_double(RP& p)
{
	int id = 14;
	{
		test_Types_delete_all_tables(p);	
		test_float<double>(p, id);
	}
}

void test_Types_crosstype_string(RP& p) {
	test_Types_delete_all_tables(p);	

	// insert a huge numeric string into the numeric field
	std::string val = "123456789012345678901234567890";
	RP::Commands::SQLExecute insert(p, "insert into fatalmind_numeric30 (id, val) values (?, ?)");
	insert.bind(0, 1);
	insert.bind(1, val);
	p.execute(insert);

	// we try to fetch it as string again
	std::string out2;
	RP::Commands::SQLSelectRow get(p, "select val from fatalmind_numeric30 where id = ?");
	get.bind(0,1);
	get.bindout(0, out2);
	p.execute(get);

	assertEqual(out2, "123456789012345678901234567890"); // string comparision


	RP::Commands::SQLExecute insert_float(p, "insert into fatalmind_double (id, val) values (?, ?)");
	insert_float.bind(0,1);
	insert_float.bind(1,"123456.789");
	p.execute(insert_float);

	RP::Commands::SQLSelectRow get_float(p, "select val from fatalmind_double where id = ?");
	get_float.bind(0,1);
	get_float.bindout(0, out2);
	p.execute(get_float);

	assertEqual(out2, "123456.789"); // string comparision

	RP::Commands::SQLSelectRow get_graf(p, "select id, ? from fatalmind_double where id=?");
	std::string teststring("sadfasfqwiohuf824cn3yt v90qw8m yru98ydc8wduy809saeyert89cwam8ty 98w4y98 cwya vafv");
	get_graf.bind(0, teststring);
	get_graf.bind(1, 1);
	int testint;
	get_graf.bindout(0, testint);
	get_graf.bindout(1, out2);
	p.execute(get_graf);
	assertEqual(out2, teststring);
}

void test_Types_string_crosstype(RP& p) {
	test_Types_delete_all_tables(p);	

	// we insert a numeric string into a string column
	std::string val = "1234";
	RP::Commands::SQLExecute insert(p, "insert into fatalmind_str (id, val) values (?, ?)");
	insert.bind(0, 1);
	insert.bind(1, val);
	p.execute(insert);

	// we try to fetch it as int
	int out;
	RP::Commands::SQLSelectRow get(p, "select val from fatalmind_str where id = ?");
	get.bind(0, 1);
	get.bindout(0, out);
	p.execute(get);

	assertEqual(out, 1234); // numeric comparision

	// we insert numerically into a string column
	int in2 = 4321;
	insert.bind(0, 2);
	insert.bind(1, in2);
	p.execute(insert);

	// we try to fetch it as string
	std::string out2;
	get.bind(0,2);
	get.bindout(0, out2);
	p.execute(get);

	assertEqual(out2, "4321"); // numeric comparision
}

void test_Types_string(RP& p) 
{
	test_Types_delete_all_tables(p);	
	std::string val;
	for (int i = 0; i < 255; ++i) {
		val.append(1, char(i%256));
	}
	RP::Commands::SQLExecute insert(p, "insert into fatalmind_str (id, val) values (?, ?)");
	insert.bind(0, 1);
	insert.bind(1, val);
	p.execute(insert);

	std::string out;
	RP::Commands::SQLSelectRow get(p, "select val from fatalmind_str where id = ?");
	get.bind(0, 1);
	get.bindout(0, out);
	p.execute(get);

	assertEqual(out, val);
}


void priv_test_lob_row(RP& p, size_t lenght, int id = 1, int offset = 0, bool a = true) {
	std::string val;
	for (size_t i = 0; i < lenght; ++i) {
		val.push_back(i%256);
	}
	RP::Commands::SQLExecute insert(p, "insert into fatalmind_lob (id, val) values (?, ?)");
	insert.bind(0, id);
	insert.bind(1, val);
	p.execute(insert);

	std::string out;
	RP::Commands::SQLSelectRow get(p, "select val from fatalmind_lob where id = ?");
	get.bind(0, id);
	get.bindout(0, out);
	p.execute(get);

	if (a) {
		assertEqual(out, val);
	}
}
void test_Types_lob(RP& p) 
{
	test_Types_delete_all_tables(p);	
	priv_test_lob_row(p, 4096, 1);
	priv_test_lob_row(p, 65535, 2);
	priv_test_lob_row(p, 1048568, 3);
}

void priv_test_perf_Types_lob_4k(RP& p, int length) 
{
	test_Types_delete_all_tables(p);	
	for (int i = 0; i < 5; ++i) {
		priv_test_lob_row(p, length, i, 0, false);
	}
}

void test_perf_Types_lob_4k(RP& p) {
	priv_test_perf_Types_lob_4k(p, 4096);
}

void test_perf_Types_lob_64k(RP& p) {
	priv_test_perf_Types_lob_4k(p, 65536);
}

void test_perf_Types_lob_1024k(RP& p) {
	priv_test_perf_Types_lob_4k(p, 1048568);
}

void test_Types_cleanup(RP& p) 
{
	test_Types_delete_all_tables(p);
}
