//********************************************************************
//*** ResourcePool/oracle/OracleExceptionFactory.cc
//*** Copyright (c) 2003-2009 by Markus Winand <mws@fatalmind.com>
//*** $Id: OracleExceptionFactory.cc,v 1.22 2009-03-18 06:54:09 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include <string.h>

#ifndef RESOURCEPOOL_ORACLE_ORACLEEXCEPTIONFACTORY_HH
#include "OracleExceptionFactory.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCE_HH
#include "Resource.hh"
#endif

namespace fatalmind {

void OracleExceptionFactory::throwit(const std::string& op
	, const std::string& fac
	, const int errornumber
	, const std::string& msg
	, const SQL::SQLStatement& sqlstm
) {
	std::ostringstream os;
	os << "ResourcePool/Error(" << op << "): " 
		<< fac << "-" << errornumber << ": " << msg;
	throw NFOOracleException(os.str(), setErrorCode(errornumber), errornumber, sqlstm);
}

void OracleExceptionFactory::throwit(const std::string& op
	, const oracle::OCIEnvHandle& env
	, sword ret
	, const SQL::SQLStatement& sqlstm
) {
	return throwit(op, env.get(), ret, OCI_HTYPE_ENV, sqlstm);
}

void OracleExceptionFactory::throwit(const std::string& op
	, const oracle::OCIErrorHandle& err
	, sword ret
	, const SQL::SQLStatement& sqlstm
) {
	return throwit(op, err.get(), ret, OCI_HTYPE_ERROR, sqlstm);
}

void OracleExceptionFactory::throwit(const std::string& op
	, const oracleHandles& hand
	, sword ret
	, const SQL::SQLStatement& sqlstm
) {
	return throwit(op, hand.getError(), ret, OCI_HTYPE_ERROR, sqlstm);
}

void OracleExceptionFactory::throwit(const std::string& oper
	, const dvoid* errhp
	, sword ret
	, ub4 type
	, const SQL::SQLStatement& sqlstm)
{
	OracleExceptionFactory oef(oper, errhp, ret, type, sqlstm);
	oef.throwit();
}


OracleExceptionFactory::OracleExceptionFactory(const std::string& oper
	, const OCIError* errhp
	, sword ret
	, const SQL::SQLStatement& sqlstm
): state()
 , msg()
 , sqlstm(sqlstm)
 , errornumber()
 , errorcode()
{
	setFromOCI(oper, errhp, OCI_HTYPE_ERROR ,ret);
	errorcode = setErrorCode(errornumber);
}

void OracleExceptionFactory::setFromOCI(const std::string& oper, const dvoid* errhp, ub4 type, sword ret)
{
	switch (ret) {
		case OCI_NEED_DATA:
			msg = "OCI_NEED_DATA";
			break;
		case OCI_NO_DATA:
			msg = "OCI_NO_DATA";
			errornumber = 999999;
			break;
		case OCI_SUCCESS_WITH_INFO:
		case OCI_ERROR: {
			OraText msgbuf[1024];

			OCIErrorGet((dvoid*)errhp
				, (ub4) 1
				, (OraText *) NULL
				, &errornumber
				, msgbuf
				, (ub4) sizeof(msgbuf)
				, (ub4) type
			);
	
			{ // called chomp in perl
				int len = strlen((char*)msgbuf);
				if ((len > 1) && (msgbuf[len-1] == '\n')) {
					msgbuf[len-1] = '\0';
				}
			}

			std::ostringstream os;
			os << "OCI_ERROR(" << oper << "): " << msgbuf;
			msg = os.str();
			}
			break;
		case OCI_INVALID_HANDLE:
			{
			std::ostringstream os;
			os << oper << ": " << "OCI_INVALID_HANDLE";
			msg = os.str();
			}
			break;
		case OCI_STILL_EXECUTING:
			msg = "OCI_STILL_EXECUTING";
			break;
		case OCI_CONTINUE:
			msg = "OCI_CONTINUE";
			break;
	};
	state = OracleException::mapORAtoXOPEN(errornumber);
}


OracleExceptionFactory::OracleExceptionFactory(const std::string& oper
	, const dvoid* errhp
	, sword ret
	, ub4 type
	, const SQL::SQLStatement& sqlstm
): state()
 , msg()
 , sqlstm(sqlstm)
 , errornumber()
 , errorcode()
{
	setFromOCI(oper, errhp, type, ret);
	errorcode = setErrorCode(errornumber);
}

void OracleExceptionFactory::throwit() const {
	// don't failover if (throw NFO..)
	if ((!state.compare(0,2, "08") && state != "57014") || errornumber == 1406) {
		throw OracleException(msg, setErrorCode(errornumber), errornumber, sqlstm);
	} else {
		throw NFOOracleException(msg, setErrorCode(errornumber), errornumber, sqlstm);
	}
}

const std::string OracleExceptionFactory::getErrorString() const {
	return msg;
}

// i did not find the place where to query o 'ORA' part with OCI.
// however, i believe (hoped) that all errors reported by OCI will have ORA-
// prefix so that always putting ORA- is correct.
std::string OracleExceptionFactory::setErrorCode(int ec) {
	std::ostringstream os;
	os << "ORA-" << std::setw(5) << std::setfill('0') << ec;
	return os.str();
}

} // namespace fatalmind

